/*
 * sushi-media-bin.css
 * Based on ekn-media-bin.css from:
 * https://github.com/endlessm/eos-knowledge-lib/tree/master/lib/eosknowledgeprivate
 *
 * Copyright (C) 2016 Endless Mobile, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Author: Juan Pablo Ugarte <ugarte@endlessm.com>
 *
 */

@define-color audio-bg-color    #4c4c4c;

sushi-media-bin {
  background: black;
}

sushi-media-bin box.overlay-bar.top {
  padding: 12px;
}

sushi-media-bin label.title {
  font: 18px Sans;
}

/* Controls */
sushi-media-bin box.overlay-bar.bottom {
    padding: 13px;
    border-radius: 5px;
}

sushi-media-bin box.overlay-bar.bottom label {
    margin: 0px 6px;
}

/* Audio nodes */
sushi-media-bin box.audio > box {
  background: @audio-bg-color;
}

sushi-media-bin box.audio > scale {
  margin: 0;
  padding: 1px 0px;
  background-image: linear-gradient(to top, @audio-bg-color 3px, transparent 3px);
}

sushi-media-bin box.audio > scale trough {
  padding: 1px 0px;
  background: #707070;
}

/* Setup key bindings */
@binding-set smb-binding-set {
  bind "space"     { "toggle" ("playback") };
  bind "f"         { "toggle" ("fullscreen") };
  bind "i"         { "toggle" ("show-stream-info") };
  bind "Right"     { "seek" (5) };
  bind "Left"      { "seek" (-5) };
  bind "Up"        { "seek" (60) };
  bind "Down"      { "seek" (-60) };
  bind "Page_Up"   { "seek" (300) };
  bind "Page_Down" { "seek" (-300) };
  bind "Home"      { "seek" (0) };
}

@binding-set smb-fullscreen-binding-set {
  bind "Escape" { "toggle" ("fullscreen") };
}

sushi-media-bin {
  -gtk-key-bindings: smb-binding-set;
}

sushi-media-bin.fullscreen {
  -gtk-key-bindings: smb-binding-set, smb-fullscreen-binding-set;
}
